#! /bin/bash

# Copyright (C) 2015 Charles Atkinson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

# Purpose:
#   Blue Light's nagios plug-in to check apt on Debian and Ubuntu

# Usage:
#   See usage.fun or use -h option.

# Versions:
#   * bash 4.2.37, 4.3.30
#   * Debian 6 Squeeze, 7 Wheezy, 8 Jessie
#   * Ubuntu 12.04 Precise, 14.04 Trusty

# Programmers' notes: error and trap handling:
#   * All errors are fatal and finalise() is called.
#   * At any time, a trapped event may transfer control to finalise().
#   * To allow finalise() to tidy up before exiting, changes that need to be
#     undone are noted with global variables named <change name>_flag and the
#     data required to undo those changes is kept in global variables.
#   * finalise() uses the same functions to undo the changes as are used when
#     they are undone routinely.
#   * $finalising_flag is used to prevent recursive calls when errors are
#     encountered while finalise() is running,

# Programmers' notes: variable names and values
#   * Directory names: *_dir.  Their values should not have a trailing /
#   * File names: *_fn
#   * Logicals: *_flag containing values $true or $false.
#     The $true and $false values allow "if [[ $debugging_flag ]]"
#   * $buf is a localised scratch buffer

# Programmers' notes: file descriptor usage
#   0 (stdin)- not changed
#   1 (stdout)- not changed except when logging request when redirected to log
#   2 (stderr)- not changed except when logging request when redirected to log

# Programmers' notes: top level function call tree
#    +
#    |
#    +-- initialise
#    |
#    +-- get_data_debian
#    |
#    +-- check_data_ubuntu
#    |
#    +-- finalise
#
# Utility functions called from various places:
#    ck_uint msg

# General function definitions in alphabetical order

#--------------------------
# Name: finalise
# Purpose: cleans up and exits
# Arguments:
#    $1: return value
#--------------------------
function finalise {
    local my_retval=$1

    # Kill the self-limiting timer process
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    [[ ${self_limiting_pid:-} != '' ]] && kill $self_limiting_pid

    exit $my_retval
}  # end of function finalise

#--------------------------
# Name: get_data_debian
# Purpose: gets apt data on Debian systems
# Arguments: none
# Global variables set:
#   n_upgrades: Number of available upgrades
#   n_security_upgrades: Number of available security upgrades
#--------------------------
function get_data_debian {
    local buf rc

    buf=$(apt-get -s dist-upgrade 2>&1)
    rc=$?
    ((rc!=0)) && msg U "Non-zero rc from apt-get.  Output: $(echo -n "$buf" | tr '[[:space:]]' ' ')"
    n_upgrades=$(echo "$buf" | grep ^Inst | wc -l)
    if ((n_upgrades==0)); then
        n_security_upgrades=0
    else
        n_security_upgrades=$(echo "$buf" | grep '^Inst.* Debian-Security:' | wc -l)
    fi

}  # end of function get_data_debian

#--------------------------
# Name: get_data_ubuntu
# Purpose: gets apt data on Ubuntu systems
# Arguments: none
# Global variables set:
#   n_upgrades: Number of available upgrades
#   n_security_upgrades: Number of available security upgrades
#--------------------------
function get_data_ubuntu {
    local buf rc
    local -r apt_check_out_regex='^[[:digit:]]+;[[:digit:]]+$'

    buf=$(/usr/lib/update-notifier/apt-check 2>&1)
    rc=$?
    buf=$(echo -n "$buf" | tr '[[:space:]]' ' ')    # Ensure single line
    if ((rc==0)); then
        [[ ! $buf =~ $apt_check_out_regex ]] \
            && msg U "Unexpected output from apt-check: $buf"
    else
       msg U "Non-zero rc $rc from apt-check.  Output: $buf"
    fi
    n_upgrades=${buf%;*}
    n_security_upgrades=${buf#*;}

}  # end of function get_data_ubuntu

#--------------------------
# Name: initialise
# Purpose: sets up environment, parses command line and sets OS-dependent variables
#--------------------------
function initialise {

    # Configure shell environment
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    export PATH=/usr/local/bin:/usr/sbin:/sbin:/usr/bin:/bin
    IFS=$' \n\t'
    set -o nounset
    shopt -s extglob            # Enable extended pattern matching operators

    # Initialise global string variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    msg_lf=$'\n    '
    script_name=${0##*/}
    script_version=0.6
    max_critical_updates=50
    max_critical_security_updates=5
    max_warn_updates=25
    max_warn_security_updates=2
    nagios_rc_ok=0
    nagios_rc_warning=1
    nagios_rc_critical=2
    nagios_rc_unknown=3

    # Initialise global logic variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    readonly false=
    readonly true=true

    # Initialise local string variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    local args=("$@")
    local emsg=
    local self_limiting_time=10
    local valid_cw_arg_regex='^[[:digit:]]+:[[:digit:]]+$'
    local uint_regex='^[[:digit:]]+$'

    # Initialise local logic variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    local opt_c_flag=$false
    local opt_t_flag=$false
    local opt_w_flag=$false

    # Parse command line
    # ~~~~~~~~~~~~~~~~~~
    while getopts :cC:dehint:UW:V opt "$@"
    do
        case $opt in
            C )
                opt_c_flag=$true
                opt_c_arg=$OPTARG
                ;;
            h )
                usage verbose
                exit 0
                ;;
            t )
                opt_t_flag=$true
                opt_t_arg=$OPTARG
                ;;
            V )
                echo "$script_name version $script_version"
                exit 0
                ;;
            W )
                opt_w_flag=$true
                opt_w_arg=$OPTARG
                ;;
            c|d|e|i|n|U )
                ;;
            : )
                emsg+=$msg_lf"Option $OPTARG must have an argument"
                ;;
            * )
                emsg+=$msg_lf"Invalid option '-$OPTARG'"
        esac
    done

    # Parse option arguments
    # ~~~~~~~~~~~~~~~~~~~~~~
    if [[ $opt_c_flag ]]; then
        if [[ $opt_c_arg =~ $valid_cw_arg_regex ]]; then
           max_critical_updates=${opt_c_arg%:*}
           max_critical_security_updates=${opt_c_arg#*:}
        else
           emsg+=$msg_lf"Invalid -c option value '$opt_c_arg' (not two :-separated uints)"
        fi
    fi
    if [[ $opt_t_flag ]]; then
        if [[ $opt_t_arg =~ $uint_regex ]]; then
           self_limiting_time=$opt_t_arg
        else
           emsg+=$msg_lf"Invalid -t option value '$opt_t_arg' (not a uint)"
        fi
    fi
    if [[ $opt_w_flag ]]; then
        if [[ $opt_w_arg =~ $valid_cw_arg_regex ]]; then
           max_warning_updates=${opt_w_arg%:*}
           max_warning_security_updates=${opt_w_arg#*:}
        else
           emsg+=$msg_lf"Invalid -w option value '$opt_w_arg' (not two :-separated uints)"
        fi
    fi

    # Test for extra arguments
    # ~~~~~~~~~~~~~~~~~~~~~~~~
    shift $(($OPTIND-1))
    if [[ $* != '' ]]; then
        emsg+=$msg_lf"Invalid extra argument(s) '$*'"
    fi

    # Report any command line errors
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if [[ $emsg != '' ]]; then
        emsg+=$msg_lf'(use -h option for help)'
        msg U "Command line error(s)$emsg"
    fi

    # Start self-limiting timer
    # ~~~~~~~~~~~~~~~~~~~~~~~~~
    trap "finalise $nagios_rc_unknown" HUP
    (
        sleep $self_limiting_time
        kill -HUP $$
        exit $nagios_rc_unknown
    )>/dev/null &
    self_limiting_pid=$!

    # Detect distro and set disto-dependent variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    local -r etc_issue=$(< /etc/issue)
    case "$etc_issue" in
        'Debian GNU/Linux 6'* )
            distro_code_name=squeeze
            ;;
        'Debian GNU/Linux 7'* )
            distro_code_name=wheezy
            ;;
        'Debian GNU/Linux 8'* )
            distro_code_name=jessie
            ;;
        'Ubuntu 12.04'* )
            distro_code_name=precise
            ;;
        'Ubuntu 14.04'* )
            distro_code_name=trusty
            ;;
    * )
            msg U "Unsupported /etc/issue content: '$etc_issue'"
    esac
}  # end of function initialise

#--------------------------
# Name: msg
# Purpose: generalised messaging interface
# Arguments:
#    $1 class: O, W, C or U indicating OK, Warning, Critical or Unknown
#    $2 message text
# Usage example:
#    msg U 'Invalid command line option'
# Output: OK messages to stdout; the rest to stderr
# Returns:
#   Does not return (calls finalise)
#--------------------------
function msg {
    local class message_text my_rc
    local prefix=APT

    # Process arguments
    # ~~~~~~~~~~~~~~~~~
    class="${1:-}"
    message_text="${2:-}"

    # Class-dependent set-up
    # ~~~~~~~~~~~~~~~~~~~~~~
    case "$class" in
        O )
            my_rc=$nagios_rc_ok
            prefix+=' OK'
            ;;
        W )
            my_rc=$nagios_rc_warning
            prefix+=' WARNING'
            ;;
        C )
            my_rc=$nagios_rc_critical
            prefix+=' CRITICAL'
            ;;
        U )
            my_rc=$nagios_rc_unknown
            prefix+=' UNKNOWN'
            ;;
        * )
            my_rc=$nagios_rc_unknown
            prefix+=' UNKNOWN'
            message_text="Programming error: msg: invalid class '$class': '$*'"
    esac

    # Write to stdout
    # ~~~~~~~~~~~~~~~
    echo "$prefix - $message_text"
    finalise $my_rc
}  #  end of function msg

#--------------------------
# Name: usage
# Purpose: prints usage message
#--------------------------
function usage {
    local msg usage

    # Build the messages
    # ~~~~~~~~~~~~~~~~~~
    usage="usage: $script_name [-C updates] [-h] [-t timeout] [-W updates] [-V] [ignored options]"
    msg='  where:'
    msg+=$'\n    -C Critical thresholds (numbers of updates:security updates)'
    msg+=$'\n       '"Default $max_critical_updates:$max_critical_security_updates"
    msg+=$'\n    -h Prints this help and exits'
    msg+=$'\n    -t Seconds before the script kills itself'". Default $self_limiting_time"
    msg+=$'\n    -W Warning thresholds (numbers of updates:security updates)'
    msg+=$'\n       '"Default $max_warn_updates:$max_warn_security_updates"
    msg+=$'\n    -V prints the version and exits'
    msg+=$'\n    -c Ignored'
    msg+=$'\n    -d Ignored'
    msg+=$'\n    -e Ignored'
    msg+=$'\n    -i Ignored'
    msg+=$'\n    -n Ignored'
    msg+=$'\n    -U Ignored'

    # Display the message(s)
    # ~~~~~~~~~~~~~~~~~~~~~~
    echo "$usage" >&2
    if [[ ${1:-} != 'verbose' ]]; then
        echo "(use -h for help)" >&2
    else
        echo "$msg" >&2
    fi
}  # end of function usage

#--------------------------
# Name: main
# Purpose: the main sequence; execution starts here
#--------------------------
initialise "${@:-}"

# Get data
# ~~~~~~~~
case $distro_code_name in
   jessie|squeeze|wheezy )
        get_data_debian
        ;;
   precise|trusty )
        get_data_ubuntu
        ;;
   * )
        msg U "Programming error: main: invalid distro code name '$distro_code_name'"
esac

# Output
# ~~~~~~
msg_class=O
if ((n_upgrades>=max_critical_updates)) \
    || ((n_security_upgrades>=max_critical_security_updates)); then
    msg_class=C
elif ((n_upgrades>=max_warn_updates)) \
    || ((n_security_upgrades>=max_warn_security_updates)); then
    msg_class=W
fi
msg $msg_class "$n_upgrades packages available for dist-upgrade ($n_security_upgrades security updates)"

# Should not get here
# ~~~~~~~~~~~~~~~~~~~
echo "$script_name: programming error: reached end of script"

finalise $nagios_rc_unknown
