#! /bin/bash

# Copyright (C) 2015 Charles Atkinson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

# Purpose:
#   Blue Light's nagios plug-in to check the cron daemon

# Usage:
#   See usage.fun or use -h option.

# Versions:
#   * bash 4.2.37, 4.3.30
#   * Debian 6 Squeeze, 7 Wheezy, 8 Jessie
#   * Ubuntu 12.04 Precise, 14.04 Trusty

# Programmers' notes: error and trap handling:
#   * All errors are fatal and finalise() is called.
#   * At any time, a trapped event may transfer control to finalise().
#   * To allow finalise() to tidy up before exiting, changes that need to be
#     undone are noted with global variables named <change name>_flag and the
#     data required to undo those changes is kept in global variables.
#   * finalise() uses the same functions to undo the changes as are used when
#     they are undone routinely.
#   * $finalising_flag is used to prevent recursive calls when errors are
#     encountered while finalise() is running,

# Programmers' notes: variable names and values
#   * Directory names: *_dir.  Their values should not have a trailing /
#   * File names: *_fn
#   * Logicals: *_flag containing values $true or $false.
#     The $true and $false values allow "if [[ $debugging_flag ]]"
#   * $buf is a localised scratch buffer

# Programmers' notes: file descriptor usage
#   0 (stdin)- not changed
#   1 (stdout)- not changed except when logging request when redirected to log
#   2 (stderr)- not changed except when logging request when redirected to log

# Programmers' notes: top level function call tree
#    +
#    |
#    +-- initialise
#    |
#    +-- get_data_systemd
#    |
#    +-- get_data_sysv_and_upstart
#    |
#    +-- finalise
#
# Utility functions called from various places:
#    ck_uint msg

# General function definitions in alphabetical order

#--------------------------
# Name: finalise
# Purpose: cleans up and exits
# Arguments:
#    $1: return value
#--------------------------
function finalise {
    local my_retval=$1

    # Kill the self-limiting timer process
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    [[ ${self_limiting_pid:-} != '' ]] && kill $self_limiting_pid

    exit $my_retval
}  # end of function finalise

#--------------------------
# Name: get_data_systemd
# Purpose: gets cron service status on systemd systems
# Arguments: none
# Global variables set:
#   msg_class: 
#       O: OK
#       W: warning
#       C: Critical
#   msg: message text
#--------------------------
function get_data_systemd {
    local buf
    local -r not_found_regex='Loaded: not-found'
    local -r running_regex='Active: active \(running\)'
    local -r stopped_regex='Active: inactive \(dead\)'

    buf=$(systemctl status cron 2>&1)
    # Because the output for a non-existent service is similar to a stopped
    # service, check for non-existent first
    if [[ $buf =~ $not_found_regex ]]; then
        msg_class=C
        msg='service cron not found'
    elif [[ $buf =~ $running_regex ]]; then
        msg_class=O
        msg='service cron running'
    elif [[ $buf =~ $stopped_regex ]]; then
        msg_class=C
        msg='service cron stopped'
    else
        msg U "Programming error: unexpected output from service cron status: '$buf'"
    fi
}  # end of function get_data_systemd

#--------------------------
# Name: get_data_sysv
# Purpose: gets cron service status on sysv systems
# Arguments: none
# Global variables set:
#   msg_class: 
#       O: OK
#       W: warning
#       C: Critical
#   msg: message text
#--------------------------
function get_data_sysv {
    local buf
    local -r not_found_regex='^cron: unrecognized service$'
    local -r running_regex='^(Checking periodic command scheduler\.*done \(running\)\.|cron is running\.)$'
    local -r stopped_regex='^(Checking periodic command scheduler\.*failed \(failed to start\)\.|cron is not running ... failed!)$'

    buf=$(service cron status 2>&1)
    if [[ $buf =~ $not_found_regex ]]; then
        msg_class=C
        msg='service cron not found'
    elif [[ $buf =~ $running_regex ]]; then
        msg_class=O
        msg='service cron running'
    elif [[ $buf =~ $stopped_regex ]]; then
        msg_class=C
        msg='service cron stopped'
    else
        msg U "Programming error: unexpected output from service cron status: '$buf'"
    fi
}  # end of function get_data_sysv

#--------------------------
# Name: get_data_upstart
# Purpose: gets cron service status on upstart systems
# Arguments: none
# Global variables set:
#   msg_class: 
#       O: OK
#       W: warning
#       C: Critical
#   msg: message text
#--------------------------
function get_data_upstart {
    local buf
    local -r not_found_regex='^cron: unrecognized service$'
    local -r running_regex='^cron start/running, process [[:digit:]]+$'
    local -r stopped_regex='^cron stop/waiting$'

    buf=$(service cron status 2>&1)
    if [[ $buf =~ $not_found_regex ]]; then
        msg_class=C
        msg='service cron not found'
    elif [[ $buf =~ $running_regex ]]; then
        msg_class=O
        msg='service cron running'
    elif [[ $buf =~ $stopped_regex ]]; then
        msg_class=C
        msg='service cron stopped'
    else
        msg U "Programming error: unexpected output from service cron status: '$buf'"
    fi
}  # end of function get_data_upstart

#--------------------------
# Name: initialise
# Purpose: sets up environment, parses command line and sets OS-dependent variables
#--------------------------
function initialise {

    # Configure shell environment
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    export PATH=/usr/local/bin:/usr/sbin:/sbin:/usr/bin:/bin
    IFS=$' \n\t'
    set -o nounset
    shopt -s extglob            # Enable extended pattern matching operators

    # Initialise global string variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    msg_lf=$'\n    '
    script_name=${0##*/}
    script_version=0.1
    nagios_rc_ok=0
    nagios_rc_warning=1
    nagios_rc_critical=2
    nagios_rc_unknown=3

    # Initialise global logic variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    readonly false=
    readonly true=true

    # Initialise local string variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    local args=("$@")
    local emsg=
    local self_limiting_time=10
    local -r uint_regex='^[[:digit:]]+$'

    # Initialise local logic variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    local opt_t_flag=$false

    # Parse command line
    # ~~~~~~~~~~~~~~~~~~
    while getopts :ht:V opt "$@"
    do
        case $opt in
            h )
                usage verbose
                exit 0
                ;;
            t )
                opt_t_flag=$true
                opt_t_arg=$OPTARG
                ;;
            V )
                echo "$script_name version $script_version"
                exit 0
                ;;
            : )
                emsg+=$msg_lf"Option $OPTARG must have an argument"
                ;;
            * )
                emsg+=$msg_lf"Invalid option '-$OPTARG'"
        esac
    done

    # Parse option arguments
    # ~~~~~~~~~~~~~~~~~~~~~~
    if [[ $opt_t_flag ]]; then
        if [[ $opt_t_arg =~ $uint_regex ]]; then
           self_limiting_time=$opt_t_arg
        else
           emsg+=$msg_lf"Invalid -t option value '$opt_t_arg' (not a uint)"
        fi
    fi

    # Test for extra arguments
    # ~~~~~~~~~~~~~~~~~~~~~~~~
    shift $(($OPTIND-1))
    if [[ $* != '' ]]; then
        emsg+=$msg_lf"Invalid extra argument(s) '$*'"
    fi

    # Report any command line errors
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if [[ $emsg != '' ]]; then
        emsg+=$msg_lf'(use -h option for help)'
        msg U "Command line error(s)$emsg"
    fi

    # Ensure running with root privilege
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ((EUID!=0)) && msg U "Must be run with effective userid 0 (root)"

    # Start self-limiting timer
    # ~~~~~~~~~~~~~~~~~~~~~~~~~
    trap "finalise $nagios_rc_unknown" HUP
    (
        sleep $self_limiting_time
        kill -HUP $$
        exit $nagios_rc_unknown
    )>/dev/null &
    self_limiting_pid=$!

    # Ensure running on a supported distro
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    local -r etc_issue=$(< /etc/issue)
    case "$etc_issue" in
        'Debian GNU/Linux 6'* )
            distro_code_name=squeeze
            ;;
        'Debian GNU/Linux 7'* )
            distro_code_name=wheezy
            ;;
        'Debian GNU/Linux 8'* )
            distro_code_name=jessie
            ;;
        'Ubuntu 12.04'* )
            distro_code_name=precise
            ;;
        'Ubuntu 14.04'* )
            distro_code_name=trusty
            ;;
    * )
            msg U "Unsupported /etc/issue content: '$etc_issue'"
    esac

    # Detect init system
    # ~~~~~~~~~~~~~~~~~~
    # Ref: http://unix.stackexchange.com/questions/18209/detect-init-system-using-the-shell
    # There may not be a general and robust algorithm.  This is good enough for
    # the small number of distros and releases supported.  It has often had to
    # be revised.
    init_system=
    buf=$(ls -l /proc/1/exe 2>&1)
    if [[ $buf = *upstart* ]]; then
        init_system=upstart
    elif [[ $buf = *systemd* ]]; then
        init_system=systemd
    elif [[ $buf = */sbin/init* ]]; then
        local -r sysv_regex='/(dev|run)/initctl'
        local -r upstart_regex=upstart
        buf=$(ls -l /proc/1/fd 2>&1)
        if [[ $buf =~ $upstart_regex ]]; then
            init_system=upstart
        elif [[ $buf =~ $sysv_regex ]]; then
            init_system=sysv
        else
            buf=$(/sbin/init --version 2>&1)
            [[ $buf = *systemd* ]] && init_system=systemd
            [[ $buf = *upstart* ]] && init_system=upstart
        fi
    fi
    [[ $init_system = '' ]] \
        && msg U "Unrecognised init system fingerprint. buf: $buf"

}  # end of function initialise

#--------------------------
# Name: msg
# Purpose: generalised messaging interface
# Arguments:
#    $1 class: O, W, C or U indicating OK, Warning, Critical or Unknown
#    $2 message text
# Usage example:
#    msg U 'Invalid command line option'
# Output: OK messages to stdout; the rest to stderr
# Returns:
#   Does not return (calls finalise)
#--------------------------
function msg {
    local class message_text my_rc
    local prefix=CRON

    # Process arguments
    # ~~~~~~~~~~~~~~~~~
    class="${1:-}"
    message_text="${2:-}"

    # Class-dependent set-up
    # ~~~~~~~~~~~~~~~~~~~~~~
    case "$class" in
        O )
            my_rc=$nagios_rc_ok
            prefix+=' OK'
            ;;
        W )
            my_rc=$nagios_rc_warning
            prefix+=' WARNING'
            ;;
        C )
            my_rc=$nagios_rc_critical
            prefix+=' CRITICAL'
            ;;
        U )
            my_rc=$nagios_rc_unknown
            prefix+=' UNKNOWN'
            ;;
        * )
            my_rc=$nagios_rc_unknown
            prefix+=' UNKNOWN'
            message_text="Programming error: msg: invalid class '$class': '$*'"
    esac

    # Write to stdout
    # ~~~~~~~~~~~~~~~
    echo "$prefix - $message_text"
    finalise $my_rc
}  #  end of function msg

#--------------------------
# Name: usage
# Purpose: prints usage message
#--------------------------
function usage {
    local msg usage

    # Build the messages
    # ~~~~~~~~~~~~~~~~~~
    usage="usage: $script_name [-h] [-t timeout] [-V]"
    msg='  where:'
    msg+=$'\n    -h Prints this help and exits'
    msg+=$'\n    -t Seconds before the script kills itself'". Default $self_limiting_time"
    msg+=$'\n    -V prints the version and exits'

    # Display the message(s)
    # ~~~~~~~~~~~~~~~~~~~~~~
    echo "$usage" >&2
    if [[ ${1:-} != 'verbose' ]]; then
        echo "(use -h for help)" >&2
    else
        echo "$msg" >&2
    fi
}  # end of function usage

#--------------------------
# Name: main
# Purpose: the main sequence; execution starts here
#--------------------------
initialise "${@:-}"

# Get data
# ~~~~~~~~
case $init_system in
   systemd )
        get_data_systemd
        ;;
   sysv )
        get_data_sysv
        ;;
   upstart )
        get_data_upstart
        ;;
   * )
        msg U "Programming error: main: init system '$init_system' does not match initialise function"
esac

# Output
# ~~~~~~
msg $msg_class "$msg"

# Should not get here
# ~~~~~~~~~~~~~~~~~~~
echo "$script_name: programming error: reached end of script"

finalise $nagios_rc_unknown
